@@ -25,14 +25,18 @@ type Job struct {
 	Total                int
 	Running              bool
 	RunningJob           bool
+	Paused               bool
 	Count403             int
 	Count429             int
 	Error                string
+	Rate                 *RateThrottle
 	startTime            time.Time
 	startTimeJob         time.Time
 	queuejobs            []QueueJob
 	queuepos             int
+	skipQueue            bool
 	currentDepth         int
+	pauseWg              sync.WaitGroup
 }
 
 type QueueJob struct {
@@ -40,17 +44,21 @@ type QueueJob struct {
 	depth int
 }
 
-func NewJob(conf *Config) Job {
+func NewJob(conf *Config) *Job {
 	var j Job
+	j.Config = conf
 	j.Counter = 0
 	j.ErrorCounter = 0
 	j.SpuriousErrorCounter = 0
 	j.Running = false
 	j.RunningJob = false
+	j.Paused = false
 	j.queuepos = 0
 	j.queuejobs = make([]QueueJob, 0)
 	j.currentDepth = 0
-	return j
+	j.Rate = NewRateThrottle(conf)
+	j.skipQueue = false
+	return &j
 }
 
 //incError increments the error counter
@@ -82,6 +90,17 @@ func (j *Job) resetSpuriousErrors() {
 	j.SpuriousErrorCounter = 0
 }
 
+//DeleteQueueItem deletes a recursion job from the queue by its index in the slice
+func (j *Job) DeleteQueueItem(index int) {
+	index = j.queuepos + index - 1
+	j.queuejobs = append(j.queuejobs[:index], j.queuejobs[index+1:]...)
+}
+
+//QueuedJobs returns the slice of queued recursive jobs
+func (j *Job) QueuedJobs() []QueueJob {
+	return j.queuejobs[j.queuepos-1:]
+}
+
 //Start the execution of the Job
 func (j *Job) Start() {
 	if j.startTime.IsZero() {
@@ -104,26 +123,32 @@ func (j *Job) Start() {
 	j.interruptMonitor()
 	for j.jobsInQueue() {
 		j.prepareQueueJob()
-
-		if j.queuepos > 1 && !j.RunningJob {
-			// Print info for queued recursive jobs
-			j.Output.Info(fmt.Sprintf("Scanning: %s", j.Config.Url))
-		}
-		j.Input.Reset()
-		j.startTimeJob = time.Now()
+		j.Reset(true)
 		j.RunningJob = true
-		j.Counter = 0
 		j.startExecution()
 	}
 
-	j.Output.Finalize()
+	err := j.Output.Finalize()
+	if err != nil {
+		j.Output.Error(err.Error())
+	}
+}
+
+// Reset resets the counters and wordlist position for a job
+func (j *Job) Reset(cycle bool) {
+	j.Input.Reset()
+	j.Counter = 0
+	j.skipQueue = false
+	j.startTimeJob = time.Now()
+	if cycle {
+		j.Output.Cycle()
+	} else {
+		j.Output.Reset()
+	}
 }
 
 func (j *Job) jobsInQueue() bool {
-	if j.queuepos < len(j.queuejobs) {
-		return true
-	}
-	return false
+	return j.queuepos < len(j.queuejobs)
 }
 
 func (j *Job) prepareQueueJob() {
@@ -132,14 +157,61 @@ func (j *Job) prepareQueueJob() {
 	j.queuepos += 1
 }
 
+//SkipQueue allows to skip the current job and advance to the next queued recursion job
+func (j *Job) SkipQueue() {
+	j.skipQueue = true
+}
+
+func (j *Job) sleepIfNeeded() {
+	var sleepDuration time.Duration
+	if j.Config.Delay.HasDelay {
+		if j.Config.Delay.IsRange {
+			sTime := j.Config.Delay.Min + rand.Float64()*(j.Config.Delay.Max-j.Config.Delay.Min)
+			sleepDuration = time.Duration(sTime * 1000)
+		} else {
+			sleepDuration = time.Duration(j.Config.Delay.Min * 1000)
+		}
+		sleepDuration = sleepDuration * time.Millisecond
+	}
+	// makes the sleep cancellable by context
+	select {
+	case <-j.Config.Context.Done(): // cancelled
+	case <-time.After(sleepDuration): // sleep
+	}
+}
+
+// Pause pauses the job process
+func (j *Job) Pause() {
+	if !j.Paused {
+		j.Paused = true
+		j.pauseWg.Add(1)
+		j.Output.Info("------ PAUSING ------")
+	}
+}
+
+// Resume resumes the job process
+func (j *Job) Resume() {
+	if j.Paused {
+		j.Paused = false
+		j.Output.Info("------ RESUMING -----")
+		j.pauseWg.Done()
+	}
+}
+
 func (j *Job) startExecution() {
 	var wg sync.WaitGroup
 	wg.Add(1)
-	go j.runProgress(&wg)
+	go j.runBackgroundTasks(&wg)
+
+	// Print the base URL when starting a new recursion queue job
+	if j.queuepos > 1 {
+		j.Output.Info(fmt.Sprintf("Starting queued job on target: %s", j.Config.Url))
+	}
+
 	//Limiter blocks after reaching the buffer, ensuring limited concurrency
 	limiter := make(chan bool, j.Config.Threads)
 
-	for j.Input.Next() {
+	for j.Input.Next() && !j.skipQueue {
 		// Check if we should stop the process
 		j.CheckStop()
 
@@ -147,26 +219,22 @@ func (j *Job) startExecution() {
 			defer j.Output.Warning(j.Error)
 			break
 		}
-
+		j.pauseWg.Wait()
 		limiter <- true
 		nextInput := j.Input.Value()
 		nextPosition := j.Input.Position()
 		wg.Add(1)
 		j.Counter++
+
 		go func() {
 			defer func() { <-limiter }()
 			defer wg.Done()
+			threadStart := time.Now()
 			j.runTask(nextInput, nextPosition, false)
-			if j.Config.Delay.HasDelay {
-				var sleepDurationMS time.Duration
-				if j.Config.Delay.IsRange {
-					sTime := j.Config.Delay.Min + rand.Float64()*(j.Config.Delay.Max-j.Config.Delay.Min)
-					sleepDurationMS = time.Duration(sTime * 1000)
-				} else {
-					sleepDurationMS = time.Duration(j.Config.Delay.Min * 1000)
-				}
-				time.Sleep(sleepDurationMS * time.Millisecond)
-			}
+			j.sleepIfNeeded()
+			j.Rate.Throttle()
+			threadEnd := time.Now()
+			j.Rate.Tick(threadStart, threadEnd)
 		}()
 
 		if !j.RunningJob {
@@ -176,38 +244,40 @@ func (j *Job) startExecution() {
 	}
 	wg.Wait()
 	j.updateProgress()
-	return
 }
 
 func (j *Job) interruptMonitor() {
 	sigChan := make(chan os.Signal, 2)
 	signal.Notify(sigChan, os.Interrupt, syscall.SIGTERM)
 	go func() {
-		for _ = range sigChan {
+		for range sigChan {
 			j.Error = "Caught keyboard interrupt (Ctrl-C)\n"
+			// resume if paused
+			if j.Paused {
+				j.pauseWg.Done()
+			}
+			// Stop the job
 			j.Stop()
 		}
 	}()
 }
 
-func (j *Job) runProgress(wg *sync.WaitGroup) {
+func (j *Job) runBackgroundTasks(wg *sync.WaitGroup) {
 	defer wg.Done()
 	totalProgress := j.Input.Total()
-	for j.Counter <= totalProgress {
-
+	for j.Counter <= totalProgress && !j.skipQueue {
+		j.pauseWg.Wait()
 		if !j.Running {
 			break
 		}
-
 		j.updateProgress()
 		if j.Counter == totalProgress {
 			return
 		}
-
 		if !j.RunningJob {
 			return
 		}
-
+		j.Rate.Adjust()
 		time.Sleep(time.Millisecond * time.Duration(j.Config.ProgressFrequency))
 	}
 }
@@ -217,6 +287,7 @@ func (j *Job) updateProgress() {
 		StartedAt:  j.startTimeJob,
 		ReqCount:   j.Counter,
 		ReqTotal:   j.Input.Total(),
+		ReqSec:     j.Rate.CurrentRate(),
 		QueuePos:   j.queuepos,
 		QueueTotal: len(j.queuejobs),
 		ErrorCount: j.ErrorCounter,
@@ -281,10 +352,8 @@ func (j *Job) runTask(input map[string][]byte, position int, retried bool) {
 	}
 	if j.Config.StopOnAll {
 		// increment 429 counter if the response code is 429
-		if j.Config.StopOnAll {
-			if resp.StatusCode == 429 {
-				j.inc429()
-			}
+		if resp.StatusCode == 429 {
+			j.inc429()
 		}
 	}
 	if j.isMatch(resp) {
@@ -303,23 +372,39 @@ func (j *Job) runTask(input map[string][]byte, position int, retried bool) {
 		j.Output.Result(resp)
 		// Refresh the progress indicator as we printed something out
 		j.updateProgress()
+		if j.Config.Recursion && j.Config.RecursionStrategy == "greedy" {
+			j.handleGreedyRecursionJob(resp)
+		}
 	}
 
-	if j.Config.Recursion && len(resp.GetRedirectLocation(false)) > 0 {
-		j.handleRecursionJob(resp)
+	if j.Config.Recursion && j.Config.RecursionStrategy == "default" && len(resp.GetRedirectLocation(false)) > 0 {
+		j.handleDefaultRecursionJob(resp)
 	}
-	return
 }
 
-//handleRecursionJob adds a new recursion job to the job queue if a new directory is found
-func (j *Job) handleRecursionJob(resp Response) {
+//handleGreedyRecursionJob adds a recursion job to the queue if the maximum depth has not been reached
+func (j *Job) handleGreedyRecursionJob(resp Response) {
+	// Handle greedy recursion strategy. Match has been determined before calling handleRecursionJob
+	if j.Config.RecursionDepth == 0 || j.currentDepth < j.Config.RecursionDepth {
+		recUrl := resp.Request.Url + "/" + "FUZZ"
+		newJob := QueueJob{Url: recUrl, depth: j.currentDepth + 1}
+		j.queuejobs = append(j.queuejobs, newJob)
+		j.Output.Info(fmt.Sprintf("Adding a new job to the queue: %s", recUrl))
+	} else {
+		j.Output.Warning(fmt.Sprintf("Maximum recursion depth reached. Ignoring: %s", resp.Request.Url))
+	}
+}
+
+//handleDefaultRecursionJob adds a new recursion job to the job queue if a new directory is found and maximum depth has
+//not been reached
+func (j *Job) handleDefaultRecursionJob(resp Response) {
+	recUrl := resp.Request.Url + "/" + "FUZZ"
 	if (resp.Request.Url + "/") != resp.GetRedirectLocation(true) {
 		// Not a directory, return early
 		return
 	}
 	if j.Config.RecursionDepth == 0 || j.currentDepth < j.Config.RecursionDepth {
 		// We have yet to reach the maximum recursion depth
-		recUrl := resp.Request.Url + "/" + "FUZZ"
 		newJob := QueueJob{Url: recUrl, depth: j.currentDepth + 1}
 		j.queuejobs = append(j.queuejobs, newJob)
 		j.Output.Info(fmt.Sprintf("Adding a new job to the queue: %s", recUrl))
@@ -343,7 +428,7 @@ func (j *Job) CalibrateResponses() ([]Response, error) {
 
 	results := make([]Response, 0)
 	for _, input := range cInputs {
-		inputs := make(map[string][]byte, 0)
+		inputs := make(map[string][]byte, len(j.Config.InputProviders))
 		for _, v := range j.Config.InputProviders {
 			inputs[v.Keyword] = []byte(input)
 		}
@@ -396,7 +481,7 @@ func (j *Job) CheckStop() {
 
 	// Check for runtime of entire process
 	if j.Config.MaxTime > 0 {
-		dur := time.Now().Sub(j.startTime)
+		dur := time.Since(j.startTime)
 		runningSecs := int(dur / time.Second)
 		if runningSecs >= j.Config.MaxTime {
 			j.Error = "Maximum running time for entire process reached, exiting."
@@ -406,7 +491,7 @@ func (j *Job) CheckStop() {
 
 	// Check for runtime of current job
 	if j.Config.MaxTimeJob > 0 {
-		dur := time.Now().Sub(j.startTimeJob)
+		dur := time.Since(j.startTimeJob)
 		runningSecs := int(dur / time.Second)
 		if runningSecs >= j.Config.MaxTimeJob {
 			j.Error = "Maximum running time for this job reached, continuing with next job if one exists."
@@ -419,11 +504,10 @@ func (j *Job) CheckStop() {
 //Stop the execution of the Job
 func (j *Job) Stop() {
 	j.Running = false
-	return
+	j.Config.Cancel()
 }
 
 //Stop current, resume to next
 func (j *Job) Next() {
 	j.RunningJob = false
-	return
 }
