@@ -7,6 +7,7 @@ import (
 	"os"
 	"path"
 	"strconv"
+	"strings"
 	"time"
 
 	"github.com/ffuf/ffuf/pkg/ffuf"
@@ -25,33 +26,22 @@ const (
 )
 
 type Stdoutput struct {
-	config  *ffuf.Config
-	Results []Result
-}
-
-type Result struct {
-	Input            map[string][]byte `json:"input"`
-	Position         int               `json:"position"`
-	StatusCode       int64             `json:"status"`
-	ContentLength    int64             `json:"length"`
-	ContentWords     int64             `json:"words"`
-	ContentLines     int64             `json:"lines"`
-	RedirectLocation string            `json:"redirectlocation"`
-	Url              string            `json:"url"`
-	ResultFile       string            `json:"resultfile"`
-	Host             string            `json:"host"`
-	HTMLColor        string            `json:"-"`
+	config         *ffuf.Config
+	Results        []ffuf.Result
+	CurrentResults []ffuf.Result
 }
 
 func NewStdoutput(conf *ffuf.Config) *Stdoutput {
 	var outp Stdoutput
 	outp.config = conf
-	outp.Results = []Result{}
+	outp.Results = make([]ffuf.Result, 0)
+	outp.CurrentResults = make([]ffuf.Result, 0)
 	return &outp
 }
 
-func (s *Stdoutput) Banner() error {
-	fmt.Printf("%s\n       v%s\n%s\n\n", BANNER_HEADER, ffuf.VERSION, BANNER_SEP)
+func (s *Stdoutput) Banner() {
+	version := strings.ReplaceAll(ffuf.Version(), "<3", fmt.Sprintf("%s<3%s", ANSI_RED, ANSI_CLEAR))
+	fmt.Fprintf(os.Stderr, "%s\n       v%s\n%s\n\n", BANNER_HEADER, version, BANNER_SEP)
 	printOption([]byte("Method"), []byte(s.config.Method))
 	printOption([]byte("URL"), []byte(s.config.Url))
 
@@ -107,12 +97,10 @@ func (s *Stdoutput) Banner() error {
 
 	// Proxies
 	if len(s.config.ProxyURL) > 0 {
-		proxy := fmt.Sprintf("%s", s.config.ProxyURL)
-		printOption([]byte("Proxy"), []byte(proxy))
+		printOption([]byte("Proxy"), []byte(s.config.ProxyURL))
 	}
 	if len(s.config.ReplayProxyURL) > 0 {
-		replayproxy := fmt.Sprintf("%s", s.config.ReplayProxyURL)
-		printOption([]byte("ReplayProxy"), []byte(replayproxy))
+		printOption([]byte("ReplayProxy"), []byte(s.config.ReplayProxyURL))
 	}
 
 	// Timeout
@@ -136,14 +124,34 @@ func (s *Stdoutput) Banner() error {
 
 	// Print matchers
 	for _, f := range s.config.Matchers {
-		printOption([]byte("Matcher"), []byte(f.Repr()))
+		printOption([]byte("Matcher"), []byte(f.ReprVerbose()))
 	}
 	// Print filters
 	for _, f := range s.config.Filters {
-		printOption([]byte("Filter"), []byte(f.Repr()))
+		printOption([]byte("Filter"), []byte(f.ReprVerbose()))
 	}
-	fmt.Printf("%s\n\n", BANNER_SEP)
-	return nil
+	fmt.Fprintf(os.Stderr, "%s\n\n", BANNER_SEP)
+}
+
+// Reset resets the result slice
+func (s *Stdoutput) Reset() {
+	s.CurrentResults = make([]ffuf.Result, 0)
+}
+
+// Cycle moves the CurrentResults to Results and resets the results slice
+func (s *Stdoutput) Cycle() {
+	s.Results = append(s.Results, s.CurrentResults...)
+	s.Reset()
+}
+
+// GetResults returns the result slice
+func (s *Stdoutput) GetCurrentResults() []ffuf.Result {
+	return s.CurrentResults
+}
+
+// SetResults sets the result slice
+func (s *Stdoutput) SetCurrentResults(results []ffuf.Result) {
+	s.CurrentResults = results
 }
 
 func (s *Stdoutput) Progress(status ffuf.Progress) {
@@ -152,11 +160,11 @@ func (s *Stdoutput) Progress(status ffuf.Progress) {
 		return
 	}
 
-	dur := time.Now().Sub(status.StartedAt)
+	dur := time.Since(status.StartedAt)
 	runningSecs := int(dur / time.Second)
-	var reqRate int
+	var reqRate int64
 	if runningSecs > 0 {
-		reqRate = int(status.ReqCount / runningSecs)
+		reqRate = status.ReqSec
 	} else {
 		reqRate = 0
 	}
@@ -167,7 +175,7 @@ func (s *Stdoutput) Progress(status ffuf.Progress) {
 	dur -= mins * time.Minute
 	secs := dur / time.Second
 
-	fmt.Fprintf(os.Stderr, "%s:: Progress: [%d/%d] :: Job [%d/%d] :: %d req/sec :: Duration: [%d:%02d:%02d] :: Errors: %d ::", TERMINAL_CLEAR_LINE, status.ReqCount, status.ReqTotal, status.QueuePos, status.QueueTotal, reqRate, hours, mins, secs, status.ErrorCount)
+	fmt.Fprintf(os.Stderr, "%s:: Progress: [%d/%d] :: Job [%d/%d] :: %d req/sec :: Duration: [%d:%02d:%02d] :: Errors: %d ::", TERMINAL_CLEAR_LINE, status.ReqCount, status.ReqTotal, status.QueuePos, status.QueueTotal, reqRate, hours, mins, secs, status.ErrorCount)
 }
 
 func (s *Stdoutput) Info(infostring string) {
@@ -175,9 +183,9 @@ func (s *Stdoutput) Info(infostring string) {
 		fmt.Fprintf(os.Stderr, "%s", infostring)
 	} else {
 		if !s.config.Colors {
-			fmt.Fprintf(os.Stderr, "%s[INFO] %s\n", TERMINAL_CLEAR_LINE, infostring)
+			fmt.Fprintf(os.Stderr, "%s[INFO] %s\n\n", TERMINAL_CLEAR_LINE, infostring)
 		} else {
-			fmt.Fprintf(os.Stderr, "%s[%sINFO%s] %s\n", TERMINAL_CLEAR_LINE, ANSI_BLUE, ANSI_CLEAR, infostring)
+			fmt.Fprintf(os.Stderr, "%s[%sINFO%s] %s\n\n", TERMINAL_CLEAR_LINE, ANSI_BLUE, ANSI_CLEAR, infostring)
 		}
 	}
 }
@@ -199,80 +207,97 @@ func (s *Stdoutput) Warning(warnstring string) {
 		fmt.Fprintf(os.Stderr, "%s", warnstring)
 	} else {
 		if !s.config.Colors {
-			fmt.Fprintf(os.Stderr, "%s[WARN] %s", TERMINAL_CLEAR_LINE, warnstring)
+			fmt.Fprintf(os.Stderr, "%s[WARN] %s\n", TERMINAL_CLEAR_LINE, warnstring)
 		} else {
 			fmt.Fprintf(os.Stderr, "%s[%sWARN%s] %s\n", TERMINAL_CLEAR_LINE, ANSI_RED, ANSI_CLEAR, warnstring)
 		}
 	}
 }
 
-func (s *Stdoutput) writeToAll(config *ffuf.Config, res []Result) error {
+func (s *Stdoutput) Raw(output string) {
+	fmt.Fprintf(os.Stderr, "%s%s", TERMINAL_CLEAR_LINE, output)
+}
+
+func (s *Stdoutput) writeToAll(filename string, config *ffuf.Config, res []ffuf.Result) error {
 	var err error
 	var BaseFilename string = s.config.OutputFile
 
 	// Go through each type of write, adding
 	// the suffix to each output file.
 
+	if config.OutputCreateEmptyFile && (len(res) == 0) {
+		return nil
+	}
+
 	s.config.OutputFile = BaseFilename + ".json"
-	err = writeJSON(s.config, s.Results)
+	err = writeJSON(filename, s.config, res)
 	if err != nil {
-		s.Error(fmt.Sprintf("%s", err))
+		s.Error(err.Error())
 	}
 
 	s.config.OutputFile = BaseFilename + ".ejson"
-	err = writeEJSON(s.config, s.Results)
+	err = writeEJSON(filename, s.config, res)
 	if err != nil {
-		s.Error(fmt.Sprintf("%s", err))
+		s.Error(err.Error())
 	}
 
 	s.config.OutputFile = BaseFilename + ".html"
-	err = writeHTML(s.config, s.Results)
+	err = writeHTML(filename, s.config, res)
 	if err != nil {
-		s.Error(fmt.Sprintf("%s", err))
+		s.Error(err.Error())
 	}
 
 	s.config.OutputFile = BaseFilename + ".md"
-	err = writeMarkdown(s.config, s.Results)
+	err = writeMarkdown(filename, s.config, res)
 	if err != nil {
-		s.Error(fmt.Sprintf("%s", err))
+		s.Error(err.Error())
 	}
 
 	s.config.OutputFile = BaseFilename + ".csv"
-	err = writeCSV(s.config, s.Results, false)
+	err = writeCSV(filename, s.config, res, false)
 	if err != nil {
-		s.Error(fmt.Sprintf("%s", err))
+		s.Error(err.Error())
 	}
 
 	s.config.OutputFile = BaseFilename + ".ecsv"
-	err = writeCSV(s.config, s.Results, true)
+	err = writeCSV(filename, s.config, res, true)
 	if err != nil {
-		s.Error(fmt.Sprintf("%s", err))
+		s.Error(err.Error())
 	}
 
 	return nil
 
 }
 
+// SaveFile saves the current results to a file of a given type
+func (s *Stdoutput) SaveFile(filename, format string) error {
+	var err error
+	switch format {
+	case "all":
+		err = s.writeToAll(filename, s.config, append(s.Results, s.CurrentResults...))
+	case "json":
+		err = writeJSON(filename, s.config, append(s.Results, s.CurrentResults...))
+	case "ejson":
+		err = writeEJSON(filename, s.config, append(s.Results, s.CurrentResults...))
+	case "html":
+		err = writeHTML(filename, s.config, append(s.Results, s.CurrentResults...))
+	case "md":
+		err = writeMarkdown(filename, s.config, append(s.Results, s.CurrentResults...))
+	case "csv":
+		err = writeCSV(filename, s.config, append(s.Results, s.CurrentResults...), false)
+	case "ecsv":
+		err = writeCSV(filename, s.config, append(s.Results, s.CurrentResults...), true)
+	}
+	return err
+}
+
+// Finalize gets run after all the ffuf jobs are completed
 func (s *Stdoutput) Finalize() error {
 	var err error
 	if s.config.OutputFile != "" {
-		if s.config.OutputFormat == "all" {
-			err = s.writeToAll(s.config, s.Results)
-		} else if s.config.OutputFormat == "json" {
-			err = writeJSON(s.config, s.Results)
-		} else if s.config.OutputFormat == "ejson" {
-			err = writeEJSON(s.config, s.Results)
-		} else if s.config.OutputFormat == "html" {
-			err = writeHTML(s.config, s.Results)
-		} else if s.config.OutputFormat == "md" {
-			err = writeMarkdown(s.config, s.Results)
-		} else if s.config.OutputFormat == "csv" {
-			err = writeCSV(s.config, s.Results, false)
-		} else if s.config.OutputFormat == "ecsv" {
-			err = writeCSV(s.config, s.Results, true)
-		}
+		err = s.SaveFile(s.config.OutputFile, s.config.OutputFormat)
 		if err != nil {
-			s.Error(fmt.Sprintf("%s", err))
+			s.Error(err.Error())
 		}
 	}
 	fmt.Fprintf(os.Stderr, "\n")
@@ -284,39 +309,37 @@ func (s *Stdoutput) Result(resp ffuf.Response) {
 	if len(s.config.OutputDirectory) > 0 {
 		resp.ResultFile = s.writeResultToFile(resp)
 	}
-	// Output the result
-	s.printResult(resp)
-	// Check if we need the data later
-	if s.config.OutputFile != "" {
-		// No need to store results if we're not going to use them later
-		inputs := make(map[string][]byte, 0)
-		for k, v := range resp.Request.Input {
-			inputs[k] = v
-		}
-		sResult := Result{
-			Input:            inputs,
-			Position:         resp.Request.Position,
-			StatusCode:       resp.StatusCode,
-			ContentLength:    resp.ContentLength,
-			ContentWords:     resp.ContentWords,
-			ContentLines:     resp.ContentLines,
-			RedirectLocation: resp.GetRedirectLocation(false),
-			Url:              resp.Request.Url,
-			ResultFile:       resp.ResultFile,
-			Host:             resp.Request.Host,
-		}
-		s.Results = append(s.Results, sResult)
+
+	inputs := make(map[string][]byte, len(resp.Request.Input))
+	for k, v := range resp.Request.Input {
+		inputs[k] = v
 	}
+	sResult := ffuf.Result{
+		Input:            inputs,
+		Position:         resp.Request.Position,
+		StatusCode:       resp.StatusCode,
+		ContentLength:    resp.ContentLength,
+		ContentWords:     resp.ContentWords,
+		ContentLines:     resp.ContentLines,
+		ContentType:      resp.ContentType,
+		RedirectLocation: resp.GetRedirectLocation(false),
+		Url:              resp.Request.Url,
+		ResultFile:       resp.ResultFile,
+		Host:             resp.Request.Host,
+	}
+	s.CurrentResults = append(s.CurrentResults, sResult)
+	// Output the result
+	s.PrintResult(sResult)
 }
 
 func (s *Stdoutput) writeResultToFile(resp ffuf.Response) string {
 	var fileContent, fileName, filePath string
 	// Create directory if needed
 	if s.config.OutputDirectory != "" {
-		err := os.Mkdir(s.config.OutputDirectory, 0750)
+		err := os.MkdirAll(s.config.OutputDirectory, 0750)
 		if err != nil {
 			if !os.IsExist(err) {
-				s.Error(fmt.Sprintf("%s", err))
+				s.Error(err.Error())
 				return ""
 			}
 		}
@@ -329,40 +352,40 @@ func (s *Stdoutput) writeResultToFile(resp ffuf.Response) string {
 	filePath = path.Join(s.config.OutputDirectory, fileName)
 	err := ioutil.WriteFile(filePath, []byte(fileContent), 0640)
 	if err != nil {
-		s.Error(fmt.Sprintf("%s", err))
+		s.Error(err.Error())
 	}
 	return fileName
 }
 
-func (s *Stdoutput) printResult(resp ffuf.Response) {
+func (s *Stdoutput) PrintResult(res ffuf.Result) {
 	if s.config.Quiet {
-		s.resultQuiet(resp)
+		s.resultQuiet(res)
 	} else {
-		if len(resp.Request.Input) > 1 || s.config.Verbose || len(s.config.OutputDirectory) > 0 {
+		if len(res.Input) > 1 || s.config.Verbose || len(s.config.OutputDirectory) > 0 {
 			// Print a multi-line result (when using multiple input keywords and wordlists)
-			s.resultMultiline(resp)
+			s.resultMultiline(res)
 		} else {
-			s.resultNormal(resp)
+			s.resultNormal(res)
 		}
 	}
 }
 
-func (s *Stdoutput) prepareInputsOneLine(resp ffuf.Response) string {
+func (s *Stdoutput) prepareInputsOneLine(res ffuf.Result) string {
 	inputs := ""
-	if len(resp.Request.Input) > 1 {
-		for k, v := range resp.Request.Input {
+	if len(res.Input) > 1 {
+		for k, v := range res.Input {
 			if inSlice(k, s.config.CommandKeywords) {
 				// If we're using external command for input, display the position instead of input
-				inputs = fmt.Sprintf("%s%s : %s ", inputs, k, strconv.Itoa(resp.Request.Position))
+				inputs = fmt.Sprintf("%s%s : %s ", inputs, k, strconv.Itoa(res.Position))
 			} else {
 				inputs = fmt.Sprintf("%s%s : %s ", inputs, k, v)
 			}
 		}
 	} else {
-		for k, v := range resp.Request.Input {
+		for k, v := range res.Input {
 			if inSlice(k, s.config.CommandKeywords) {
 				// If we're using external command for input, display the position instead of input
-				inputs = strconv.Itoa(resp.Request.Position)
+				inputs = strconv.Itoa(res.Position)
 			} else {
 				inputs = string(v)
 			}
@@ -371,30 +394,30 @@ func (s *Stdoutput) prepareInputsOneLine(resp ffuf.Response) string {
 	return inputs
 }
 
-func (s *Stdoutput) resultQuiet(resp ffuf.Response) {
-	fmt.Println(s.prepareInputsOneLine(resp))
+func (s *Stdoutput) resultQuiet(res ffuf.Result) {
+	fmt.Println(s.prepareInputsOneLine(res))
 }
 
-func (s *Stdoutput) resultMultiline(resp ffuf.Response) {
+func (s *Stdoutput) resultMultiline(res ffuf.Result) {
 	var res_hdr, res_str string
 	res_str = "%s%s    * %s: %s\n"
-	res_hdr = fmt.Sprintf("%s[Status: %d, Size: %d, Words: %d, Lines: %d]", TERMINAL_CLEAR_LINE, resp.StatusCode, resp.ContentLength, resp.ContentWords, resp.ContentLines)
-	res_hdr = s.colorize(res_hdr, resp.StatusCode)
+	res_hdr = fmt.Sprintf("%s[Status: %d, Size: %d, Words: %d, Lines: %d]", TERMINAL_CLEAR_LINE, res.StatusCode, res.ContentLength, res.ContentWords, res.ContentLines)
+	res_hdr = s.colorize(res_hdr, res.StatusCode)
 	reslines := ""
 	if s.config.Verbose {
-		reslines = fmt.Sprintf("%s%s| URL | %s\n", reslines, TERMINAL_CLEAR_LINE, resp.Request.Url)
-		redirectLocation := resp.GetRedirectLocation(false)
+		reslines = fmt.Sprintf("%s%s| URL | %s\n", reslines, TERMINAL_CLEAR_LINE, res.Url)
+		redirectLocation := res.RedirectLocation
 		if redirectLocation != "" {
 			reslines = fmt.Sprintf("%s%s| --> | %s\n", reslines, TERMINAL_CLEAR_LINE, redirectLocation)
 		}
 	}
-	if resp.ResultFile != "" {
-		reslines = fmt.Sprintf("%s%s| RES | %s\n", reslines, TERMINAL_CLEAR_LINE, resp.ResultFile)
+	if res.ResultFile != "" {
+		reslines = fmt.Sprintf("%s%s| RES | %s\n", reslines, TERMINAL_CLEAR_LINE, res.ResultFile)
 	}
-	for k, v := range resp.Request.Input {
+	for k, v := range res.Input {
 		if inSlice(k, s.config.CommandKeywords) {
 			// If we're using external command for input, display the position instead of input
-			reslines = fmt.Sprintf(res_str, reslines, TERMINAL_CLEAR_LINE, k, strconv.Itoa(resp.Request.Position))
+			reslines = fmt.Sprintf(res_str, reslines, TERMINAL_CLEAR_LINE, k, strconv.Itoa(res.Position))
 		} else {
 			// Wordlist input
 			reslines = fmt.Sprintf(res_str, reslines, TERMINAL_CLEAR_LINE, k, v)
@@ -403,15 +426,14 @@ func (s *Stdoutput) resultMultiline(resp ffuf.Response) {
 	fmt.Printf("%s\n%s\n", res_hdr, reslines)
 }
 
-func (s *Stdoutput) resultNormal(resp ffuf.Response) {
-	var res_str string
-	res_str = fmt.Sprintf("%s%-23s [Status: %s, Size: %d, Words: %d, Lines: %d]", TERMINAL_CLEAR_LINE, s.prepareInputsOneLine(resp), s.colorize(fmt.Sprintf("%d", resp.StatusCode), resp.StatusCode), resp.ContentLength, resp.ContentWords, resp.ContentLines)
-	fmt.Println(res_str)
+func (s *Stdoutput) resultNormal(res ffuf.Result) {
+	resnormal := fmt.Sprintf("%s%-23s [Status: %s, Size: %d, Words: %d, Lines: %d]", TERMINAL_CLEAR_LINE, s.prepareInputsOneLine(res), s.colorize(fmt.Sprintf("%d", res.StatusCode), res.StatusCode), res.ContentLength, res.ContentWords, res.ContentLines)
+	fmt.Println(resnormal)
 }
 
 func (s *Stdoutput) colorize(input string, status int64) string {
 	if !s.config.Colors {
-		return fmt.Sprintf("%s", input)
+		return input
 	}
 	colorCode := ANSI_CLEAR
 	if status >= 200 && status < 300 {
@@ -430,7 +452,7 @@ func (s *Stdoutput) colorize(input string, status int64) string {
 }
 
 func printOption(name []byte, value []byte) {
-	fmt.Printf(" :: %-16s : %s\n", name, value)
+	fmt.Fprintf(os.Stderr, " :: %-16s : %s\n", name, value)
 }
 
 func inSlice(key string, slice []string) bool {
